# -*- coding: utf-8 -*-
"""IsatData Pro modem constants.

This module provides mapping of constants used within an IDP modem.

"""

from enum import Enum, IntEnum, IntFlag


class MessagePriority(IntEnum):
    NONE = 0
    HIGH = 1
    MEDH = 2
    MEDL = 3
    LOW = 4


class DataFormat(IntEnum):
    TEXT = 1
    HEX = 2
    BASE64 = 3


CONTROL_STATES = {
    0: 'Stopped',
    1: 'Waiting for GNSS fix',
    2: 'Starting search',
    3: 'Beam search',
    4: 'Beam found',
    5: 'Beam acquired',
    6: 'Beam switch in progress',
    7: 'Registration in progress',
    8: 'Receive only',
    9: 'Downloading Bulletin Board',
    10: 'Active',
    11: 'Blocked',
    12: 'Confirm previously registered beam',
    13: 'Confirm requested beam',
    14: 'Connect to confirmed beam'
}


class SatlliteControlState(IntEnum):
    STOPPED = 0
    GNSS_WAIT = 1
    SEARCH_START = 2
    BEAM_SEARCH = 3
    BEAM_FOUND = 4
    BEAM_ACQUIRED = 5
    BEAM_SWITCH = 6
    REGISTERING = 7
    RECEIVE_ONLY = 8
    BB_DOWNLOAD = 9
    ACTIVE = 10
    BLOCKED = 11
    CONFIRM_PREVIOUS_BEAM = 12
    CONFIRM_REQUESTED_BEAM = 13
    CONNECT_CONFIRMED_BEAM = 14


class BeamSearchState(IntEnum):
    IDLE = 0
    SEARCH_ANY_TRAFFIC = 1
    SEARCH_LAST_TRAFFIC = 2
    RESERVED = 3
    SEARCH_NEW_TRAFFIC = 4
    SEARCH_BULLETIN_BOARD = 5
    DELAY_TRAFFIC_SEARCH = 6


class MessageState(IntEnum):
    UNAVAILABLE = 0
    RX_PENDING = 1
    RX_COMPLETE = 2
    RX_RETRIEVED = 3
    TX_READY = 4
    TX_SENDING = 5
    TX_COMPLETE = 6
    TX_FAILED = 7
    TX_CANCELLED = 8


AT_ERROR_CODES = {
    0: 'OK',
    4: 'ERROR',
    100: 'ERR_INVALID_CRC_SEQUENCE',
    101: 'ERR_UNKNOWN_COMMAND',
    102: 'ERR_INVALID_COMMAND_PARAMETERS',
    103: 'ERR_MESSAGE_LENGTH_EXCEEDS_FORMAT_SIZE',
    104: 'ERR_RESERVED_104',
    105: 'ERR_SYSTEM_ERROR',
    106: 'ERR_QUEUE_INSUFFICIENT_RESOURCES',
    107: 'ERR_MESSAGE_NAME_ALREADY_IN_USE',
    108: 'ERR_TIMEOUT_OCCURRED',
    109: 'ERR_UNAVAILABLE',
    110: 'ERR_RESERVED_110',
    111: 'ERR_RESERVED_111',
    112: 'ERR_ATTEMPT_TO_WRITE_READ_ONLY_PARAMETER'
}

WAKEUP_PERIODS = {
    0: 'SECONDS_5',
    1: 'SECONDS_30',
    2: 'MINUTES_1',
    3: 'MINUTES_3',
    4: 'MINUTES_10',
    5: 'MINUTES_30',
    6: 'MINUTES_60',
    7: 'MINUTES_2',
    8: 'MINUTES_5',
    9: 'MINUTES_15',
    10: 'MINUTES_20'
}

POWER_MODES = {
    0: 'MOBILE_POWERED',
    1: 'FIXED_POWERED',
    2: 'MOBILE_BATTERY',
    3: 'FIXED_BATTERY',
    4: 'MOBILE_MINIMAL',
    5: 'MOBILE_PARKED'
}

# Certain GNSS modes are SkyWave hardware-specific
GNSS_MODES = {
    0: 'GPS',   # HW v4
    1: 'GLONASS',   # HW v5
    2: 'BEIDOU',   # HW v5.2
    10: 'GPS+GLONASS',   # UBX-M80xx
    11: 'GPS+BEIDOU',   # UBX-M80xx
    12: 'GLONASS+BEIDOU',   # UBX-M80xx
}

# Dynamic Platform Model - use with caution
GNSS_DPM_MODES = {
    0: 'PORTABLE',
    2: 'STATIONARY',
    3: 'PEDESTRIAN',
    4: 'AUTOMOTIVE',
    5: 'SEA',
    6: 'AIR_1G',
    7: 'AIR_2G',
    8: 'AIR_4G'
}


class SignalLevelRegional(Enum):
    """Qualitative descriptors for SNR/CN0 values for a IDP Regional Beam.
    
    BARS_n: *n* is a scale from 0..5 to be used as greaterThan threshold
    NONE, MARGINAL, GOOD: a scale to be used as greaterOrEqual threshold

    """
    BARS_0 = 0
    BARS_1 = 37.0
    BARS_2 = 39.0
    BARS_3 = 41.0
    BARS_4 = 43.0
    BARS_5 = 45.5
    INVALID = 55.0


class SignalQuality(IntEnum):
    NONE = 0
    WEAK = 1
    LOW = 2
    MID = 3
    GOOD = 4
    STRONG = 5
    WARNING = 6


class EventNotification(IntFlag):
    GNSS_FIX_NEW = 0b000000000001
    MESSAGE_MT_RECEIVED = 0b000000000010
    MESSAGE_MO_COMPLETE = 0b000000000100
    NETWORK_REGISTERED = 0b000000001000
    MODEM_RESET_COMPLETE = 0b000000010000
    JAMMING_ANTENNA_CHANGE = 0b000000100000
    MODEM_RESET_PENDING = 0b000001000000
    WAKEUP_PERIOD_CHANGE = 0b000010000000
    UTC_TIME_SYNC = 0b000100000000
    GNSS_FIX_TIMEOUT = 0b001000000000
    EVENT_TRACE_CACHED = 0b010000000000
    NETWORK_PING_ACKNOWLEDGED = 0b100000000000


class TransmitterStatus(IntEnum):
    RX_ONLY_NOT_REGISTERED = 4
    OK = 5
    SUSPENDED = 6
    MUTED = 7
    BLOCKED = 8


class EventTrace:
    def __init__(self,
                 trace_class: int,
                 trace_subclass: int,
                 data: tuple) -> None:
        self.trace_class = trace_class
        self.trace_subclass = trace_subclass
        self.data = data


EVENT_TRACE_SATELLITE_GENERAL = EventTrace(
    trace_class=3,
    trace_subclass=1,
    data=(
        ('subframe_number', 'uint'),
        ('traffic_vcid', 'uint'),
        ('configuration_id', 'uint'),
        ('beam_number', 'uint'),
        ('reserved04', 'uint'),
        ('tx_access_sip', 'uint'),
        ('tx_access_operator', 'uint'),
        ('tx_access_user', 'uint'),
        ('tx_suspend_flags', {
            0x1: 'BEAM_REGISTRATION',
            0x2: 'BEAM_SWITCH',
            0x4: 'RESERVED',
            0x8: 'BLOCKED',
        }),
        ('tx_messages_active', 'uint'),
        ('tx_messages_total', 'uint'),
        ('tx_packets_active', 'uint'),
        ('tx_state', {
            0: 'ACTIVE',
            1: 'SUSPENDING',
            2: 'SUSPENDED_PENDING_GNSS'}),
        ('active_rx_messages', 'uint'),
        ('beamswitch_averaging_window', 'uint'),
        ('beamswitch_averaging_count', 'uint'),
        ('c_n_x100', 'uint'),
        ('beamsample_threshold', 'uint'),
        ('beamsample_timer', 'uint'),
        ('flags', {
            0x1: 'REGISTERED',
            0x2: 'SENDING_BEAM_REGISTRATION',
            0x10: 'BEAM_SEARCH',
            0x20: 'BEAM_SAMPLE_REQUIRED',
            0x40: 'BEAM_SWITCH_PENDING',
            0x100: 'GNSS_VALID',
            0x200: 'GNSS_REQUIRED',
            0x400: 'GNSS_PENDING',
        }),
        ('gnss_state_timer', 'uint'),
        ('reserved21', 'uint'),
        ('satellite_control_state', SatlliteControlState),
        ('beam_search_state', BeamSearchState),
    )
)


EVENT_TRACES = (
    EVENT_TRACE_SATELLITE_GENERAL,
)


GEOBEAMS = {
  1: 'AMER RB1',
  2: 'AMER RB2',
  3: 'AMER RB3',
  4: 'AMER RB4',
  5: 'AMER RB5',
  6: 'AMER RB6',
  7: 'AMER RB7',
  8: 'AMER RB8',
  9: 'AMER RB9',
  10: 'AMER RB10',
  11: 'AMER RB11',
  12: 'AMER RB12',
  13: 'AMER RB13',
  14: 'AMER RB14',
  15: 'AMER RB15',
  16: 'AMER RB16',
  17: 'AMER RB17',
  18: 'AMER RB18',
  19: 'AMER RB19',
  61: 'AORW SC',
  21: 'EMEA RB1',
  22: 'EMEA RB2',
  23: 'EMEA RB3',
  24: 'EMEA RB4',
  25: 'EMEA RB5',
  26: 'EMEA RB6',
  27: 'EMEA RB7',
  28: 'EMEA RB8',
  29: 'EMEA RB9',
  30: 'EMEA RB10',
  31: 'EMEA RB11',
  32: 'EMEA RB12',
  33: 'EMEA RB13',
  34: 'EMEA RB14',
  35: 'EMEA RB15',
  36: 'EMEA RB16',
  37: 'EMEA RB17',
  38: 'EMEA RB18',
  39: 'EMEA RB19',
  41: 'APAC RB1',
  42: 'APAC RB2',
  43: 'APAC RB3',
  44: 'APAC RB4',
  45: 'APAC RB5',
  46: 'APAC RB6',
  47: 'APAC RB7',
  48: 'APAC RB8',
  49: 'APAC RB9',
  50: 'APAC RB10',
  51: 'APAC RB11',
  52: 'APAC RB12',
  53: 'APAC RB13',
  54: 'APAC RB14',
  55: 'APAC RB15',
  56: 'APAC RB16',
  57: 'APAC RB17',
  58: 'APAC RB18',
  59: 'APAC RB19',
  90: 'MEAS RB10',
  91: 'MEAS RB11',
  92: 'MEAS RB12',
  93: 'MEAS RB15',
}



"""AT Command protocol factory for pyserial-asyncio.

**WARNING**: This module is untested as a work in progress refactor of the
legacy atcommand_async module.

"""

import asyncio
import logging
from time import time

import serial_asyncio
from idpmodem.aterror import AtCrcError, AtTimeout
from idpmodem.crcxmodem import get_crc, validate_crc

_log = logging.getLogger(__name__)


class AtProtocol(asyncio.Protocol):
    
    TERMINATOR = '\r'
    ENCODING = 'utf-8'
    UNICODE_HANDLING = 'replace'
    DEFAULT_TIMEOUT = 5

    def __init__(self,
                 crc: bool = False,
                 event_callback: callable = None,
                 at_timeout: int = 5,
                 ):
        super().__init__()
        self.transport: asyncio.Transport = None
        self.crc = crc
        self.alive = True
        self.at_timeout = at_timeout
        self.pending_command = None
        self.command_time = None
        self.response_time = None
        self.response: 'list[str]' = []
        self.response_complete = False
        self.filter = None
        self._debug = False
        self.event_callback = event_callback

    def connection_made(self, transport) -> None:
        self.transport = transport
        _log.debug('Serial AT protocol connection opened')
    
    def connection_lost(self, exc) -> None:
        _log.warning('Serial AT procotol connection lost')
        self.transport.loop.stop()
        self.transport = None
        if isinstance(exc, Exception):
            raise exc

    def data_received(self, data: bytes) -> None:
        self.buffer.extend(data)
        if self.pending_command is not None:
            if data == b'\r':
                #: Echo case
                if self.buffer == self.pending_command.encode() + b'\r':
                    echo = self.buffer
                    #: reset buffer for response
                    self.buffer = bytearray(b'')
                    self.handle_packet(echo)
            elif data == b'\n':
                if (self.buffer != bytearray(b'\r\n')
                    and self.buffer != bytearray(b'\n')):
                    #: Framed/multiline response, error code or CRC
                    packet = self.buffer
                    self.buffer = bytearray(b'')
                    self.handle_packet(packet)
                elif self.buffer == bytearray(b'\n'):
                    #: (Unexpected) drop any empty lines
                    self.buffer = bytearray(b'')
        else:   # no pending command
            if data == b'\n':
                unsolicited = self.buffer
                self.buffer = bytearray(b'')
                self.handle_packet(unsolicited)

    def handle_packet(self, packet: bytearray):
        """Decodes packet(s) - inherited from LineReader."""
        self.handle_line(packet.decode(self.ENCODING, self.UNICODE_HANDLING))

    def handle_line(self, line: str):
        """Enqueues lines for parsing by command handler.

        Args:
            line: The unicode string received from the serial port.
        """
        if self.pending_command is not None:
            self.handle_response(line)
        else:
            if line != '\n':
                self.handle_event(line)

    def write_line(self, line: str):
        """Appends a terminator, encodes and writes text to the transport."""
        send: str = line + self.TERMINATOR
        self.transport.write(send.encode(self.ENCODING, self.UNICODE_HANDLING))
    
    def handle_event(self, unsolicited: str):
        """Calls a user-defined function with the unicode string.

        Args:
            unsolicited: A unicode string terminated by <lf>.

        """
        if unsolicited is not None:
            if self.event_callback is not None:
                self.event_callback(unsolicited)
            else: 
                unsolicited = unsolicited.replace('\r', '<cr>')
                unsolicited = unsolicited.replace('\n', '<lf>')
                _log.warning(f'Unhandled event: {unsolicited}')

    def handle_response(self, line: str):
        content = line.strip()
        if self.response_time is None:
            self.response_time = time()
        if content == self.pending_command:
            pass   # ignore echo
        elif content == 'OK':
            self.response.append(line)
            if '%CRC=0' in self.pending_command:
                self.crc = False
            if (not self.crc and '%CRC=1' not in self.pending_command):
                # No need to wait for CRC
                self.response_complete = True
        elif content.startswith('*'):
            if not self.crc:
                if not '%CRC=1' in self.pending_command:
                    _log.warning('Inferring CRC enabled')
                self.crc = True
            crc = content.replace('*', '')
            if not validate_crc(''.join(self.response), crc):
                raise AtCrcError(f'INVALID_CRC_RESPONSE')
            self.response_complete = True
        else:   #: including 'ERROR'
            self.response.append(line)
            # keep parsing in case CRC follows    

    def _clean_response(self,
                        lines: 'list[str]',
                        filter: 'list[str]' = [],
                        debug: bool = False,
                        ) -> 'list[str]':
        """Removes empty lines from response and returns a list.
        
        Args:
            lines: A list of reponse lines.
            filter: Optional list of strings/substrings to filter from response.
            debug: If True, logs the command latency
        
        Returns:
            List with filtered and stripped lines

        Raises:
            ValueError if filter is invalid.

        """
        if filter and not isinstance(filter, list):
            raise ValueError('filter must be a list of strings')
        if debug:
            latency = round(self.response_time - self.command_time, 3)
            _log.debug(f'Command {self.pending_command}'
                          f' latency: {latency} seconds')
        for l in range(len(lines)):
            lines[l] = lines[l].strip()
            if isinstance(filter, list) and len(filter) > 0:
                for s in filter:
                    if lines[l].startswith(s):
                        s += ':' if lines[l].startswith(f'{s}:') else ''
                        lines[l] = lines[l].replace(s, '').strip()
        return [x for x in lines if x != '']

    async def command(self,
                command: str,
                filter: 'list[str]' = [],
                timeout: int = 5,
                debug: bool = False,
                ) -> 'list[str]':
        """Send an AT command and wait for the response.

        Returns the response as a list.  If an error response code was
        received then 'ERROR' will be the only string in the list.

        .. todo: generalize for OK-only response, and provide error detail

        Args:
            command: The AT command
            timeout: Time to wait for response in seconds (default 5)
        
        Returns:
            A list of strings. The list will be ['ERROR'] in case of a problem.

        Raises:
            AtCrcError if CRC does not match.
            AtTimeout if the request timed out.

        """
        if self.pending_command is not None:
            raise 
        timeout = 1 if timeout < 1 else timeout
        command = get_crc(command) if self.crc else command
        self.pending_command = command
        self.response_complete = False
        self.response_time = None
        self.command_time = time()
        self.write_line(command)
        try:
            await self._timeout_check(timeout=timeout)
            return self._clean_response(self.response, filter, debug)
        except asyncio.TimeoutError:
            raise AtTimeout(f'TIMEOUT ({command})')
        finally:
            self._cleanup()

    async def _timeout_check(self):
        while not self.response_complete:
            await asyncio.sleep(1)

    def _cleanup(self):
        self.pending_command = None
        self.response = []



"""Helpers for operating a headless device for example Raspberry Pi.

"""
import binascii
from string import printable

import serial.tools.list_ports as list_ports


def validate_serial_port(target: str, verbose: bool = False) -> 'bool|tuple':
    """Validates a given serial port as available on the host.

    When working with different OS and platforms, using a serial port to connect
    to a modem can be simplified by *validate_serial_port*.

    If target port is not found, a list of available ports is returned.
    Labels known FTDI and Prolific serial/USB drivers.

    Args:
        target: Target port name e.g. ``/dev/ttyUSB0``
    
    Returns:
        True or False if detail is False
        (valid: bool, description: str) if detail is True
    """
    found = False
    detail = ''
    ser_ports = [tuple(port) for port in list(list_ports.comports())]
    for port in ser_ports:
        if target == port[0]:
            found = True
            usb_id = str(port[2])
            if 'USB VID:PID=0403:6001' in usb_id:
                driver = 'Serial FTDI FT232 (RS485/RS422/RS232)'
            elif 'USB VID:PID=067B:2303' in usb_id:
                driver = 'Serial Prolific PL2303 (RS232)'
            else:
                driver = 'Serial vendor/device {}'.format(usb_id)
            detail = '{} on {}'.format(driver, port[0])
    if not found and len(ser_ports) > 0:
        for port in ser_ports:
            if len(detail) > 0:
                detail += ','
            detail += " {}".format(port[0])
        detail = 'Available ports:' + detail
    return (found, detail) if verbose else found

def is_hex_string(s: str) -> bool:
    """Returns True if the string consists exclusively of hexadecimal chars."""
    hex_chars = '0123456789abcdefABCDEF'
    return all(c in hex_chars for c in s)


def bytearray_to_str(arr: bytearray) -> str:
    """Converts a bytearray to a readable text string."""
    s = ''
    for b in bytearray(arr):
        if chr(b) in printable:
            s += chr(b)
        else:
            s += '{0:#04x}'.format(b).replace('0x', '\\')
    return s


def bytearray_to_hex_str(arr: bytearray) -> str:
    """Converts a bytearray to a hex string."""
    return binascii.hexlify(bytearray(arr)).decode()


def bytearray_to_b64_str(arr: bytearray) -> str:
    """Converts a bytearray to a base64 string."""
    return binascii.b2a_base64(bytearray(arr)).strip().decode()



"""Utilities for validating and parsing NMEA data into a `Location` object.

"""
import logging
from calendar import timegm
from datetime import datetime
from functools import reduce
from operator import xor

_log = logging.getLogger(__name__)


class NmeaException(Exception):
    """An exception triggered during NMEA sentence parsing."""
    pass


class NmeaInvalid(NmeaException):
    """Invalid NMEA sentence."""
    pass


class NmeaChecksumError(NmeaException):
    """NMEA sentence checksum error."""
    pass


class GnssSatelliteInfo(object):
    """Information specific to a GNSS satellite.
    
    Attributes:
        prn: The PRN code (Pseudo-Random Number sequence)
        elevation: The satellite elevation
        azimuth: The satellite azimuth
        snr: The satellite Signal-to-Noise Ratio
    """
    def __init__(self,
                    prn: int,
                    elevation: int,
                    azimuth: int,
                    snr: int):
        """Initializes GNSS satellite details."""
        self.prn = prn
        self.elevation = elevation
        self.azimuth = azimuth
        self.snr = snr


class Location:
    """A set of location-based information for a given point in time.
    
    Uses 90.0/180.0 if latitude/longitude are unknown

    Attributes:
        latitude: decimal degrees
        longitude: decimal degrees
        altitude: in metres
        speed: in knots
        heading: in degrees
        timestamp: in seconds since 1970-01-01T00:00:00Z
        satellites: in view at time of fix
        fix_type: 1=None, 2=2D or 3=3D
        pdop: Probability Dilution of Precision
        hdop: Horizontal Dilution of Precision
        vdop: Vertical Dilution of Precision
        time_iso: ISO 8601 formatted timestamp
        satellites_info: `GnssSatelliteInfo`

    """

    def __init__(self, **kwargs):
        """Initializes a Location with default latitude/longitude 90/180."""
        self.latitude = float(kwargs.get('latitude', 90.0))
        self.longitude = float(kwargs.get('longitude', 180.0))
        self.resolution = int(kwargs.get('resolution', 6))
        self.altitude = float(kwargs.get('altitude', 0.0))   # metres
        self.speed = float(kwargs.get('speed', 0.0))  # knots
        self.heading = float(kwargs.get('heading', 0.0))   # degrees
        self.timestamp = int(kwargs.get('timestamp', 0))   # seconds (unix)
        self.satellites = int(kwargs.get('satellites', 0))
        self.fix_type = int(kwargs.get('fix_type', 1))
        self.pdop = float(kwargs.get('pdop', 99.9))
        self.hdop = float(kwargs.get('hdop', 99.9))
        self.vdop = float(kwargs.get('vdop', 99.9))
        self.satellites_info: 'list[GnssSatelliteInfo]' = kwargs.get(
            'satellites_info', []
        )

    @property
    def time_iso(self) -> str:
        return f'{datetime.utcfromtimestamp(self.timestamp).isoformat()}Z'

    def _update_satellites_info(self,
                                satellites_info: 'list[GnssSatelliteInfo]'):
        """Populates satellite information based on NMEA GSV data."""
        for satellite_info in satellites_info:
            if isinstance(satellite_info, GnssSatelliteInfo):
                new = True
                for i, info in enumerate(self.satellites_info):
                    if info.prn == satellite_info.prn:
                        new = False
                        self.satellites_info[i] = satellite_info
                        break
                if new:
                    self.satellites_info.append(satellite_info)
    
    def serialize(self, include_satellite_details: bool = False) -> str:
        d = vars(self)
        if include_satellite_details:
            for i, satellite_info in enumerate(d['satellites_info']):
                d['satellites_info'][i] = vars(satellite_info)
        else:
            del d['satellites_info']
        return d


def validate_nmea_checksum(sentence: str) -> 'tuple[bool, str]':
    """Validates NMEA sentence using checksum according to the standard.

    Args:
        sentence: NMEA sentence including checksum
    
    Returns:
       A tuple with Boolean result (checksum correct) and
       raw NMEA data string, with prefix $Gx and checksum suffix removed

    """
    sentence = sentence.strip('\n').strip('\r')
    nmeadata, cksum = sentence.split('*', 1)
    nmeadata = nmeadata.replace('$', '')
    xcksum = str("%0.2x" % (reduce(xor, (ord(c) for c in nmeadata), 0))).upper()
    return (cksum == xcksum), nmeadata[2:]


def location_from_nmea(nmea_data_set: 'list[str]',
                       degrees_resolution: int = 6,
                       ) -> Location:
    """Returns a Location object based on an NMEA sentences data set.
    
    Args:
        nmea_data_set: A list of valid NMEA sentences
        degrees_resolution: The desired precision of latitude/longitude
    
    Returns:
        A `Location` object

    """
    loc = Location()
    for sentence in nmea_data_set:
        if not sentence.startswith('$G'):
            raise NmeaInvalid('Invalid NMEA string {}'.format(sentence))
        valid, nmea_data = validate_nmea_checksum(sentence)
        if not valid:
            raise NmeaChecksumError('Invalid NMEA checksum for {}'.format(sentence))
        sentence_type = nmea_data[0:3]
        if sentence_type == 'GGA':          # GGA is essential fix information for 3D location and accuracy
            gga = nmea_data.split(',')      # $GPGGA,123519,4807.038,N,01131.000,E,1,08,0.9,545.4,M,46.9,M,,*4
            '''
            gga_utc_hhmmss = gga[1]         # Fix taken at 12:35:19 UTC
            gga_latitude_dms = gga[2]       # Latitude 48 deg 07.038'
            gga_latitude_ns = gga[3]        # Latitude N
            gga_longitude_dms = gga[4]      # Longitude 11 deg 31.000'
            gga_longitude_ew = gga[5]       # Longitude E
            gga_quality = gga[6]            # Fix quality
            gga_fix_qualities = (
                'invalid',                  # 0 = invalid
                'GPS fix',                  # 1 = GPS fix (SPS)
                'DGPS fix',                 # 2 = DGPS fix
                'PPS fix',                  # 3 = PPS fix
                'RTK',                      # 4 = Real Time Kinematic
                'Float RTK',                # 5 = Float RTK
                'Estimated',                # 6 = estimated (dead reckoning)
                'Manual',                   # 7 = Manual input mode
                'Simulation'                # 8 = Simulation mode
            )
            '''
            gga_satellites = gga[7]         # Number of satellites being tracked
            gga_hdop = gga[8]               # Horizontal dilution of precision
            gga_altitude = gga[9]           # Altitude above mean sea level
            '''
            gga_altitude_unit = gga[10]     # Altitude units (M)eters
            gga_height_wgs84 = gga[11]      # Height of geoid (mean sea level) above WGS84 ellipsoid
            gga_height_unit = gga[12]       # Height units (M)eters
            gga_dgps_update_time = gga[13]  # Time in seconds since last DGPS update
            gga_dgps_station = gga[14]      # DGPS station ID number
            '''
            satellites = int(gga_satellites)
            if loc.satellites < satellites:
                loc.satellites = satellites
            else:
                # TODO: log this case; should be limited to GPS simulation in Modem Simulator (3 satellites)
                pass
            loc.altitude = float(gga_altitude) if gga_altitude != '' else 0.0
            loc.hdop = min(float(gga_hdop), 32.0) if gga_hdop != '' else 32.0
        # RMC
        elif sentence_type == 'RMC':          # RMC Recommended Minimum is used for most location information
            rmc = nmea_data.split(',')        # $GPRMC,123519,A,4807.038,N,01131.000,E,022.4,084.4,230394,003.1,W*6A
            rmc_fixtime_utc_hhmmss = rmc[1]   # fix taken at 12:35:19 UTC
            # rmc_active = RMC[2]             # status (A)ctive or (V)oid
            rmc_latitude_dms = rmc[3]         # 4807.038 = 48 deg 07.038'
            rmc_latitude_ns = rmc[4]          # (N)orth or (S)outh
            rmc_longitude_dms = rmc[5]        # 01131.000 = 11 deg 31.000'
            rmc_longitude_ew = rmc[6]         # (E)east or (W)est
            rmc_speed_knots = rmc[7]          # 022.4 = 22.4 knots
            rmc_heading_deg_true = rmc[8]     # 084.4 = 84.4 degrees True
            rmc_date_ddmmyy = rmc[9]          # date 23rd of March 1994
            '''
            rmc_mag_var_mag = rmc[10]         # Magnetic Variation (magnitude)
            rmc_mag_var_dir = rmc[11]         # Magnetic Variation (direction)
            '''
            # Convert text values to workable numbers
            year = int(rmc_date_ddmmyy[4:6]) + 2000
            month = int(rmc_date_ddmmyy[2:4])
            day = int(rmc_date_ddmmyy[0:2])
            hour = int(rmc_fixtime_utc_hhmmss[0:2])
            minute = int(rmc_fixtime_utc_hhmmss[2:4])
            second = int(rmc_fixtime_utc_hhmmss[4:6])
            dt = datetime(year, month, day, hour, minute, second)
            loc.timestamp = int(timegm(dt.timetuple()))
            # Convert to decimal degrees latitude/longitude
            if rmc_longitude_dms != '' and rmc_longitude_dms != '':
                loc.latitude = round(float(rmc_latitude_dms[0:2]) +
                                     float(rmc_latitude_dms[2:]) /
                                     60.0, degrees_resolution)
                if rmc_latitude_ns == 'S':
                    loc.latitude *= -1
                loc.longitude = round(float(rmc_longitude_dms[0:3]) +
                                      float(rmc_longitude_dms[3:]) /
                                      60.0, degrees_resolution)
                if rmc_longitude_ew == 'W':
                    loc.longitude *= -1
            loc.speed = float(rmc_speed_knots) if rmc_speed_knots != '' else 0.0  # multiply by 1.852 for kph
            if rmc_heading_deg_true:
                loc.heading = float(rmc_heading_deg_true)
            else:
                loc.heading = 0.0
        # GSA
        elif sentence_type == 'GSA':                    # GSA is used for DOP and active satellites
            gsa = nmea_data.split(',')                  # $GPGSA,A,3,04,05,,09,12,,,24,,,,,2.5,1.3,2.1*39
            #: gsa_auto = gsa[1]                           # Auto selection of 2D or 3D fix (M = manual)
            gsa_fix_type = gsa[2]                       # 3D fix type
            '''
            gsa_fix_types = {
                'none': 1,
                '2D': 2,
                '3D': 3
            }
            '''
            gsa_prns = []                               # PRNs of satellites used for fix (space for 12)
            for prn in range(1, 12):
                gsa_prns.append(gsa[prn+2])             # offset of prns in the split array is [3]
            gsa_pdop = gsa[15]                          # Probability dilution of precision (DOP), above 20 is bad
            #: gsa_hdop = gsa[16]                          # Horizontal DOP
            gsa_vdop = gsa[17]                          # Vertical DOP
            # Use GSA for fix_type, PDOP, VDOP (HDOP comes from GGA)
            loc.fix_type = int(gsa_fix_type) if gsa_fix_type != '' else 0
            loc.pdop = min(float(gsa_pdop), 32.0) if gsa_pdop != '' else 32.0
            loc.vdop = min(float(gsa_vdop), 32.0) if gsa_vdop != '' else 32.0
        # GSV
        elif sentence_type == 'GSV':         # Satellites in View
            gsv = nmea_data.split(',')       # $GPGSV,2,1,08,01,40,083,46,02,17,308,41,12,07,344,39,14,22,228,45*75
            '''
            gsv_sentences = gsv[1]           # Number of sentences for full data
            gsv_sentence = gsv[2]            # Sentence number (up to 4 satellites per sentence)
            '''
            gsv_satellites = gsv[3]          # Number of satellites in view
            # following supports up to 4 satellites per sentence
            satellites_info = []
            if (len(gsv) - 4) % 4 > 0:
                # TODO: warn/log this case of extra GSV data in sentence
                pass
            num_satellites_in_sentence = int((len(gsv)-4)/4)
            for i in range(1, num_satellites_in_sentence+1):
                prn = int(gsv[i*4]) if gsv[i*4] != '' else 0             # satellite PRN number
                elevation = int(gsv[i*4+1]) if gsv[i*4+1] != '' else 0   # Elevation in degrees
                azimuth = int(gsv[i*4+2]) if gsv[i*4+2] != '' else 0     # Azimuth in degrees
                snr = int(gsv[i*4+3]) if gsv[i*4+3] != '' else 0         # Signal to Noise Ratio
                satellites_info.append(GnssSatelliteInfo(prn,
                                                         elevation,
                                                         azimuth,
                                                         snr))
            loc._update_satellites_info(satellites_info)
            satellites = int(gsv_satellites) if gsv_satellites != '' else 0
            if loc.satellites < satellites:
                loc.satellites = satellites
            else:
                # TODO: log this case; should be limited to GPS simulation in Modem Simulator (3 satellites)
                pass
        else:
            err = f'{sentence[0:3]} NMEA sentence type not recognized'
            raise NmeaException(err)
    return loc



"""Codec functions for IDP Common Message Format supported by Inmarsat MGS.

Also supported on ORBCOMM IGWS.

"""
import logging
import xml.etree.ElementTree as ET
from binascii import b2a_base64
from math import ceil, log2
from struct import pack, unpack
from warnings import warn
from xml.dom.minidom import parseString

from idpmodem.constants import DataFormat

__version__ = '2.0.0'
_log = logging.getLogger(__name__)


DATA_TYPES = {
    'bool': 'BooleanField',
    'int_8': 'SignedIntField',
    'uint_8': 'UnsignedIntField',
    'int_16': 'SignedIntField',
    'uint_16': 'UnsignedIntField',
    'int_32': 'SignedIntField',
    'uint_32': 'UnsignedIntField',
    'int_64': 'SignedIntField',
    'uint_64': 'UnsignedIntField',
    'float': 'DataField',
    'double': 'DataField',
    'string': 'StringField',
    'data': 'DataField',
    'enum': 'EnumField',
    'array': 'ArrayField',   # TODO: support for array type
}
XML_NAMESPACE = {
    'xsi': 'http://www.w3.org/2001/XMLSchema-instance',
    'xsd': 'http://www.w3.org/2001/XMLSchema'
}
SIN_RANGE = (16, 255)

for ns in XML_NAMESPACE:
    ET.register_namespace(ns, XML_NAMESPACE[ns])


def optimal_bits(value_range: 'tuple[int, int]') -> int:
    """Returns the optimal number of bits for encoding a specified range.
    
    Args:
        value_range: A tuple with the minimum and maximum values.
    
    Returns:
        The number of bits to optimally encode the value.
    
    Raises:
        ValueError if the 

    """
    if (not isinstance(value_range, tuple) or
        len(value_range) != 2 or
        not all(isinstance(x, int) for x in value_range) or
        value_range[0] >= value_range[1]):
        #: non-compliant
        raise ValueError('value_range must be of form (min, max)')
    total_range = value_range[1] - value_range[0]
    total_range += 1 if value_range[0] == 0 else 0
    return max(1, ceil(log2(value_range[1] - value_range[0])))


def _encode_field_length(length) -> str:
    if length < 128:
        return f'0{length:07b}'
    return f'1{length:015b}'


def _decode_field_length(binstr: str) -> 'tuple[int, int]':
    if binstr[0] == '0':
        bit_index = 8
    else:
        bit_index = 16
    length = int(binstr[1:bit_index], 2)
    return (length, bit_index)


def _attribute_equivalence(reference: object,
                           other: object,
                           exclude: "list[str]" = None) -> bool:
    for attr, val in reference.__dict__.items():
        if exclude is not None and attr in exclude:
            continue
        if not hasattr(other, attr) or val != other.__dict__[attr]:
            return False
    return True


def _indent_xml(elem, level=0):
    xmlstr = parseString(ET.tostring(elem)).toprettyxml(indent="  ")
    # i = "\n" + level*"  "
    # j = "\n" + (level-1)*"  "
    # if len(elem):
    #     if not elem.text or not elem.text.strip():
    #         elem.text = i + "  "
    #     if not elem.tail or not elem.tail.strip():
    #         elem.tail = i
    #     for subelem in elem:
    #         _indent_xml(subelem, level+1)
    #     if not elem.tail or not elem.tail.strip():
    #         elem.tail = j
    # else:
    #     if level and (not elem.tail or not elem.tail.strip()):
    #         elem.tail = j
    # return elem
    return xmlstr


class CodecList(list):
    """Base class for a specific object type list.
    
    Used for Fields, Messages, Services.

    Attributes:
        list_type: The object type the list is comprised of.

    """
    def __init__(self, list_type: object):
        super().__init__()
        self.list_type = list_type

    def add(self, obj: object) -> bool:
        """Add an object to the end of the list.

        Args:
            obj (object): A valid object according to the list_type
        
        Raises:
            ValueError if there is a duplicate or invalid name,
                invalid value_range or unsupported data_type
        """
        if not isinstance(obj, self.list_type):
            raise ValueError(f'Invalid {self.list_type} definition')
        for o in self:
            if o.name == obj.name:
                raise ValueError(f'Duplicate {self.list_type}'
                                 f' name {obj.name} found')
        self.append(obj)
        return True

    def __getitem__(self, n: 'str|int') -> object:
        """Retrieves an object by name or index.
        
        Args:
            n: The object name or list index
        
        Returns:
            object

        """
        if isinstance(n, str):
            for o in self:
                if o.name == n:
                    return o
            raise ValueError(f'{self.list_type} name {n} not found')
        return super().__getitem__(n)

    def __setitem__(self, n: 'str|int', value):
        if isinstance(n, str):
            for o in self:
                if o.name == n:
                    o.value = value
                    break
        else:
            super().__setitem__(n, value)

    def delete(self, name: str) -> bool:
        """Delete an object from the list by name.
        
        Args:
            name: The name of the object.

        Returns:
            boolean: success
        """
        for o in self:
            if o.name == name:
                self.remove(o)
                return True
        return False


class FieldCodec:
    """The base class for a Field.
    
    Attributes:
        data_type (str): The data type from a supported list.
        name (str): The unique Field name.
        description (str): Optional description.
        optional (bool): Optional indication the field is optional.

    """
    def __init__(self,
                 name: str,
                 data_type: str,
                 description: str = None,
                 optional: bool = False) -> None:
        """Instantiates the base field.
        
        Args:
            name: The field name must be unique within a Message.
            data_type: The data type represented within the field.
            description: (Optional) Description/purpose of the field.
            optional: (Optional) Indicates if the field is mandatory.
            
        """
        if data_type not in DATA_TYPES:
            raise ValueError(f'Invalid data type {data_type}')
        if name is None or name.strip() == '':
            raise ValueError('Invalid name must be non-empty')
        self.data_type = data_type
        self.name = name
        self.description = description
        self.optional = optional
    
    def __repr__(self) -> str:
        from pprint import pformat
        return pformat(vars(self), indent=4)
    
    def _base_xml(self) -> ET.Element:
        xsi_type = DATA_TYPES[self.data_type]
        xmlfield = ET.Element('Field', attrib={
            '{http://www.w3.org/2001/XMLSchema-instance}type': xsi_type
        })
        name = ET.SubElement(xmlfield, 'Name')
        name.text = self.name
        if self.description:
            description = ET.SubElement(xmlfield, 'Description')
            description.text = str(self.description)
        if self.optional:
            optional = ET.SubElement(xmlfield, 'Optional')
            optional.text = 'true'
        return xmlfield
    
    def decode(*args, **kwargs):
        """Subclass"""
        raise NotImplementedError('You must subclass and define decode')


class Fields(CodecList):
    """The list of Fields defining a Message or ArrayElement."""
    def __init__(self, fields: 'list[FieldCodec]' = None):
        super().__init__(list_type=FieldCodec)
        if fields is not None:
            for field in fields:
                self.add(field)
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Fields):
            return NotImplemented
        if len(self) != len(other):
            return False
        for field in self:
            if field != other[field.name]:
                return False
        return True


class MessageCodec:
    """The Payload structure for Message Definition Files uploaded to a Mailbox.
    
    Attributes:
        name (str): The message name
        sin (int): The Service Identification Number
        min (int): The Message Identification Number
        fields (list): A list of Fields
        description (str): Optional description
        is_forward (bool): Indicates if the message is mobile-terminated

    """

    def __init__(self,
                 name: str,
                 sin: int,
                 min: int,
                 description: str = None,
                 is_forward: bool = False,
                 fields: Fields = None):
        """Instantiates a Message.
        
        Args:
            name: The message name should be unique within the xMessages list.
            sin: The Service Identification Number (16..255)
            min: The Message Identification Number (0..255)
            description: (Optional) Description/purpose of the Message.
            is_forward: Indicates if the message is intended to be
                Mobile-Terminated.
            fields: Optional definition of fields during instantiation.

        """
        if not isinstance(sin, int) or sin not in range(16, 256):
            raise ValueError(f'Invalid SIN {sin} must be in range 16..255')
        if not isinstance(min, int) or min not in range (0, 256):
            raise ValueError(f'Invalid MIN {min} must be in range 0..255')
        self.name = name
        self.description = description
        self.is_forward = is_forward
        self.sin = sin
        self.min = min
        self.fields: Fields = fields or Fields()

    @property
    def fields(self) -> list:
        return self._fields
    
    @fields.setter
    def fields(self, fields: "list[FieldCodec]"):
        if not all(isinstance(field, FieldCodec) for field in fields):
            raise ValueError('Invalid field found in list')
        self._fields = fields

    @property
    def ota_size(self) -> int:
        ota_bits = 2 * 8
        for field in self.fields:
            ota_bits += field.bits + (1 if field.optional else 0)
        return ceil(ota_bits / 8)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MessageCodec):
            return NotImplemented
        return _attribute_equivalence(self, other)

    def decode(self, databytes: bytes) -> None:
        """Parses and stores field values from raw data (received over-the-air).
        
        Args:
            databytes: A bytes array (typically from the forward message)
        """
        binary_str = ''.join(format(int(b), '08b') for b in databytes)
        bit_offset = 16   #: Begin after SIN/MIN bytes
        for field in self.fields:
            if field.optional:
                present = binary_str[bit_offset] == '1'
                bit_offset += 1
                if not present:
                    continue
            bit_offset += field.decode(binary_str[bit_offset:])

    def derive(self, databytes: bytes) -> None:
        """Derives field values from raw data bytes (received over-the-air).
        
        Deprecated/replaced by `decode`.

        Args:
            databytes: A bytes array (typically from the forward message)

        """
        self.decode(databytes)

    def encode(self,
               data_format: int = DataFormat.BASE64,
               exclude: list = None) -> dict:
        """Encodes using the specified data format (base64 or hex).

        Args:
            data_format (int): 2=ASCII-Hex, 3=base64
            exclude (list[str]): A list of optional field names to exclude
        
        Returns:
            Dictionary with sin, min, data_format and data to pass into AT%MGRT
                or atcommand function `message_mo_send`
        """
        if data_format not in [DataFormat.BASE64, DataFormat.HEX]:
            raise ValueError(f'data_format {data_format} unsupported')
        bin_str = ''
        for field in self.fields:
            if field.optional:
                if exclude is not None and field.name in exclude:
                    present = False
                elif hasattr(field, 'value'):
                    present = field.value is not None
                elif hasattr(field, 'elements'):
                    present = field.elements is not None
                else:
                    raise ValueError('Unknown value of optional')
                bin_str += '1' if present else '0'
                if not present:
                    continue
            bin_str += field.encode()
        for _ in range(0, 8 - len(bin_str) % 8):   #:pad to next byte
            bin_str += '0'
        _format = f'0{int(len(bin_str) / 8 * 2)}X'   #:hex bytes 2 chars
        hex_str = format(int(bin_str, 2), _format)
        if (self.is_forward and len(hex_str) / 2 > 9998 or
            not self.is_forward and len(hex_str) / 2 > 6398):
            raise ValueError(f'{len(hex_str) / 2} bytes exceeds maximum size'
                             ' for Payload')
        if data_format == DataFormat.HEX:
            data = hex_str
        else:
            data = b2a_base64(bytearray.fromhex(hex_str)).strip().decode()
        return {
            'sin': self.sin,
            'min': self.min,
            'data_format': data_format,
            'data': data
        }

    def xml(self, indent: bool = False) -> ET.Element:
        """Returns the XML definition for a Message Definition File."""
        xmessage = ET.Element('Message')
        name = ET.SubElement(xmessage, 'Name')
        name.text = self.name
        min = ET.SubElement(xmessage, 'MIN')
        min.text = str(self.min)
        fields = ET.SubElement(xmessage, 'Fields')
        for field in self.fields:
            fields.append(field.xml())
        return xmessage if not indent else _indent_xml(xmessage)


class Messages(CodecList):
    """The list of Messages (Forward or Return) within a Service."""
    def __init__(self, sin: int, is_forward: bool):
        super().__init__(list_type=MessageCodec)
        self.sin = sin
        self.is_forward = is_forward
    
    def add(self, message: MessageCodec) -> bool:
        """Add a message to the list if it matches the parent SIN.

        Overrides the base class add method.

        Args:
            message (object): A valid Message
        
        Raises:
            ValueError if there is a duplicate or invalid name,
                invalid value_range or unsupported data_type

        """
        if not isinstance(message, MessageCodec):
            raise ValueError('Invalid message definition')
        if message.sin != self.sin:
            raise ValueError(f'Message SIN {message.sin} does not match'
                             f' service {self.sin}')
        for m in self:
            if m.name == message.name:
                raise ValueError(f'Duplicate message name {message.name} found')
            if m.min == message.min:
                raise ValueError(f'Duplicate message MIN {message.min} found')
        self.append(message)
        return True


class ServiceCodec:
    """A data structure holding a set of related Forward and Return Messages.
    
    Attributes:
        name (str): The service name
        sin (int): Service Identification Number or codec service id (16..255)
        description (str): A description of the service (unsupported)
        messages_forward (list): A list of mobile-terminated Message definitions
        messages_return (list): A list of mobile-originated Message definitions

    """
    def __init__(self,
                 name: str,
                 sin: int,
                 description: str = None,
                 messages_forward: "list[MessageCodec]" = None,
                 messages_return: "list[MessageCodec]" = None) -> None:
        """Instantiates a Service made up of Messages.
        
        Args:
            name: The service name should be unique within a MessageDefinitions
            sin: The Service Identification Number (16..255)
            description: (Optional)
        """
        if not isinstance(name, str) or name == '':
            raise ValueError(f'Invalid service name {name}')
        if sin not in range(16, 256):
            raise ValueError('Invalid SIN must be 16..255')
        self.name = name
        self.sin = sin
        if description is not None:
            warn('Service Description not currently supported')
        self.description = None
        self.messages_forward = messages_forward or Messages(self.sin,
                                                             is_forward=True)
        self.messages_return = messages_return or Messages(self.sin,
                                                           is_forward=False)
    
    def xml(self, indent: bool = False) -> ET.Element:
        """Returns the XML structure of the Service for a MDF."""
        if len(self.messages_forward) == 0 and len(self.messages_return) == 0:
            raise ValueError(f'No messages defined for service {self.sin}')
        xservice = ET.Element('Service')
        name = ET.SubElement(xservice, 'Name')
        name.text = str(self.name)
        sin = ET.SubElement(xservice, 'SIN')
        sin.text = str(self.sin)
        if self.description:
            desc = ET.SubElement(xservice, 'Description')
            desc.text = str(self.description)
        if len(self.messages_forward) > 0:
            forward_messages = ET.SubElement(xservice, 'ForwardMessages')
            for m in self.messages_forward:
                forward_messages.append(m.xml())
        if len(self.messages_return) > 0:
            return_messages = ET.SubElement(xservice, 'ReturnMessages')
            for m in self.messages_return:
                return_messages.append(m.xml())
        return xservice if not indent else _indent_xml(xservice)


class Services(CodecList):
    """The list of Service(s) within a MessageDefinitions."""
    def __init__(self, services: "list[ServiceCodec]" = None):
        super().__init__(list_type=ServiceCodec)
        if services is not None:
            for service in services:
                if not isinstance(service, ServiceCodec):
                    raise ValueError(f'Invalid Service {service}')
                self.add(service)
    
    def add(self, service: ServiceCodec) -> None:
        """Adds a Service to the list of Services."""
        if not isinstance(service, ServiceCodec):
            raise ValueError(f'{service} is not a valid Service')
        if service.name in self:
            raise ValueError(f'Duplicate Service {service.name}')
        for existing_service in self:
            if existing_service.sin == service.sin:
                raise ValueError(f'Duplicate SIN {service.sin}')
        self.append(service)


class BooleanField(FieldCodec):
    """A Boolean field."""
    def __init__(self,
                 name: str,
                 description: str = None,
                 optional: bool = False,
                 default: bool = False,
                 value: bool = None) -> None:
        super().__init__(name=name,
                         data_type='bool',
                         description=description,
                         optional=optional)
        """Instantiates a BooleanField.
        
        Args:
            name: The field name must be unique within a Message.
            description: An optional description/purpose for the field.
            optional: Indicates if the field is optional in the Message.
            default: A default value for the boolean.
            value: Optional value to set during initialization.

        """
        self.default = default
        self.value = value if value is not None else default
    
    @property
    def default(self):
        return self._default

    @default.setter
    def default(self, v: bool):
        if v is not None and not isinstance(v, bool):
            raise ValueError(f'Invalid boolean value {v}')
        self._default = v

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, v: bool):
        if v is not None and not isinstance(v, bool):
            raise ValueError(f'Invalid boolean value {v}')
        self._value = v

    @property
    def bits(self):
        return 1
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BooleanField):
            return NotImplemented
        return _attribute_equivalence(self, other)

    def encode(self) -> str:
        """Returns the binary string of the field value."""
        if self.value is None and not self.optional:
            raise ValueError('No value assigned to field')
        return '1' if self.value else '0'

    def decode(self, binary_str: str) -> int:
        """Populates the field value from binary and returns the next offset.
        
        Args:
            binary_str (str): The binary string to decode
        
        Returns:
            The bit offset after parsing
        """
        self.value = True if binary_str[0] == '1' else False
        return 1

    def xml(self) -> ET.Element:
        xmlfield = self._base_xml()
        if self.default:
            default = ET.SubElement(xmlfield, 'Default')
            default.text = 'true'
        return xmlfield


class EnumField(FieldCodec):
    """An enumerated field sends an index over-the-air representing a string."""
    def __init__(self,
                 name: str,
                 items: "list[str]",
                 size: int,
                 description: str = None,
                 optional: bool = False,
                 default: int = None,
                 value: int = None) -> None:
        """Instantiates a EnumField.
        
        Args:
            name: The field name must be unique within a Message.
            items: A list of strings (indexed from 0).
            size: The number of *bits* used to encode the index over-the-air.
            description: An optional description/purpose for the field.
            optional: Indicates if the field is optional in the Message.
            default: A default value for the enum.
            value: Optional value to set during initialization.

        """
        super().__init__(name=name,
                         data_type='enum',
                         description=description,
                         optional=optional)
        if items is None or not all(isinstance(item, str) for item in items):
            raise ValueError('Items must all be strings')
        if not isinstance(size, int) or size < 1:
            raise ValueError('Size must be integer greater than 0 bits')
        self.items = items
        self.size = size
        self.default = default
        self.value = value if value is not None else self.default
    
    def _validate_enum(self, v: 'int|str') -> 'int|None':
        if v is not None:
            if isinstance(v, str):
                if v not in self.items:
                    raise ValueError(f'Invalid value {v}')
                for index, item in enumerate(self.items):
                    if item == v:
                        return index
            elif isinstance(v, int):
                if v < 0 or v >= len(self.items):
                    raise ValueError(f'Invalid enum index {v}')
            else:
                raise ValueError(f'Invalid value {v}')
        return v

    @property
    def items(self):
        return self._items
    
    @items.setter
    def items(self, l: list):
        if not isinstance(l, list) or not all(isinstance(x, str) for x in l):
            raise ValueError('Items must be a list of strings')
        self._items = l

    @property
    def default(self):
        if self._default is None:
            return None
        return self.items[self._default]
    
    @default.setter
    def default(self, v: 'int|str'):
        self._default = self._validate_enum(v)

    @property
    def value(self):
        if self._value is None:
            return None
        return self.items[self._value]
    
    @value.setter
    def value(self, v: 'int|str'):
        self._value = self._validate_enum(v)

    @property
    def size(self):
        return self._size
    
    @size.setter
    def size(self, v: int):
        if not isinstance(v, int) or v < 1:
            raise ValueError('Size must be integer greater than zero')
        minimum_bits = optimal_bits((0, len(self.items)))
        if v < minimum_bits:
            raise ValueError(f'Size must be at least {minimum_bits}'
                             ' to support item count')
        self._size = v

    @property
    def bits(self):
        return self.size
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, EnumField):
            return NotImplemented
        return _attribute_equivalence(self, other)

    def encode(self) -> str:
        """Returns the binary string of the field value."""
        if self.value is None:
            raise ValueError(f'No value configured in EnumField {self.name}')
        _format = f'0{self.bits}b'
        binstr = format(self.items.index(self.value), _format)
        return binstr

    def decode(self, binary_str: str) -> int:
        """Populates the field value from binary and returns the next offset.
        
        Args:
            binary_str (str): The binary string to decode
        
        Returns:
            The bit offset after parsing
        """
        self.value = binary_str[:self.bits]
        return self.bits

    def xml(self) -> ET.Element:
        xmlfield = self._base_xml()
        size = ET.SubElement(xmlfield, 'Size')
        size.text = str(self.size)
        items = ET.SubElement(xmlfield, 'Items')
        for string in self.items:
            item = ET.SubElement(items, 'string')
            item.text = str(string)
        if self.default:
            default = ET.SubElement(xmlfield, 'Default')
            default.text = str(self.default)
        return xmlfield


class UnsignedIntField(FieldCodec):
    """An unsigned integer value using a defined number of bits over-the-air."""
    def __init__(self,
                 name: str,
                 size: int,
                 data_type: str = 'uint_16',
                 description: str = None,
                 optional: bool = False,
                 default: int = None,
                 value: int = None) -> None:
        """Instantiates a UnsignedIntField.
        
        Args:
            name: The field name must be unique within a Message.
            size: The number of *bits* used to encode the integer over-the-air
                (maximum 32).
            data_type: The integer type represented (for decoding).
            description: An optional description/purpose for the string.
            optional: Indicates if the string is optional in the Message.
            default: A default value for the string.
            value: Optional value to set during initialization.

        """
        if data_type not in ['uint_8', 'uint_16', 'uint_32']:
            raise ValueError(f'Invalid unsignedint type {data_type}')
        super().__init__(name=name,
                         data_type=data_type,
                         description=description,
                         optional=optional)
        self.size = size
        self.default = default
        self.value = value if value is not None else default
    
    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, value: int):
        if not isinstance(value, int) or value < 1:
            raise ValueError('Size must be integer greater than 0 bits')
        data_type_size = int(self.data_type.split('_')[1])
        if value > data_type_size:
            warn(f'Size {value} larger than required by {self.data_type}')
        self._size = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, v: int):
        clip = False
        if v is not None:
            if not isinstance(v, int) or v < 0:
                raise ValueError('Unsignedint must be non-negative integer')
            if v > 2**self.size - 1:
                self._value = 2**self.size - 1
                warn(f'Clipping unsignedint at max value {self._value}')
                clip = True
        if not clip:
            self._value = v
    
    @property
    def default(self):
        return self._default
    
    @default.setter
    def default(self, v: int):
        if v is not None:
            if v > 2**self.size - 1 or v < 0:
                raise ValueError(F'Invalid unsignedint default {v}')
        self._default = v
    
    @property
    def bits(self):
        return self.size
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, UnsignedIntField):
            return NotImplemented
        return _attribute_equivalence(self, other)

    def encode(self) -> str:
        """Returns the binary string of the field value."""
        if self.value is None:
            raise ValueError(f'No value defined in UnsignedIntField {self.name}')
        _format = f'0{self.bits}b'
        return format(self.value, _format)

    def decode(self, binary_str: str) -> int:
        """Populates the field value from binary and returns the next offset.
        
        Args:
            binary_str (str): The binary string to decode
        
        Returns:
            The bit offset after parsing
        """
        self.value = int(binary_str[:self.bits], 2)
        return self.bits

    def xml(self) -> ET.Element:
        xmlfield = self._base_xml()
        size = ET.SubElement(xmlfield, 'Size')
        size.text = str(self.size)
        if self.default:
            default = ET.SubElement(xmlfield, 'Default')
            default.text = str(self.default)
        return xmlfield


class SignedIntField(FieldCodec):
    """A signed integer value using a defined number of bits over-the-air."""
    def __init__(self,
                 name: str,
                 size: int,
                 data_type: str = 'int_16',
                 description: str = None,
                 optional: bool = False,
                 default: int = None,
                 value: int = None) -> None:
        """Instantiates a SignedIntField.
        
        Args:
            name: The field name must be unique within a Message.
            size: The number of *bits* used to encode the integer over-the-air
                (maximum 32).
            data_type: The integer type represented (for decoding).
            description: An optional description/purpose for the string.
            optional: Indicates if the string is optional in the Message.
            default: A default value for the string.
            value: Optional value to set during initialization.

        """
        if data_type not in ['int_8', 'int_16', 'int_32']:
            raise ValueError(f'Invalid unsignedint type {data_type}')
        super().__init__(name=name,
                         data_type=data_type,
                         description=description,
                         optional=optional)
        self.size = size
        self.default = default
        self.value = value if value is not None else default
    
    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, value: int):
        if not isinstance(value, int) or value < 1:
            raise ValueError('Size must be integer greater than 0 bits')
        self._size = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, v: int):
        clip = False
        if v is not None:
            if not isinstance(v, int):
                raise ValueError('Unsignedint must be non-negative integer')
            if v > (2**self.size / 2) - 1:
                self._value = int(2**self.size / 2) - 1
                warn(f'Clipping signedint at max value {self._value}')
                clip = True
            if v < -(2**self.size / 2):
                self._value = -1 * int(2**self.size / 2)
                warn(f'Clipping signedint at min value {self._value}')
                clip = True
        if not clip:
            self._value = v
    
    @property
    def default(self):
        return self._default
    
    @default.setter
    def default(self, v: int):
        if v is not None:
            if not isinstance(v, int):
                raise ValueError(f'Invalid signed integer {v}')
            if v > (2**self.size / 2) - 1 or v < -(2**self.size / 2):
                raise ValueError(f'Invalid default {v}')
        self._default = v
    
    @property
    def bits(self):
        return self.size
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SignedIntField):
            return NotImplemented
        return _attribute_equivalence(self, other)

    def encode(self) -> str:
        """Returns the binary string of the field value."""
        if self.value is None:
            raise ValueError(f'No value defined in UnsignedIntField {self.name}')
        _format = f'0{self.bits}b'
        if self.value < 0:
            invertedbin = format(self.value * -1, _format)
            twocomplementbin = ''
            i = 0
            while len(twocomplementbin) < len(invertedbin):
                twocomplementbin += '1' if invertedbin[i] == '0' else '0'
                i += 1
            binstr = format(int(twocomplementbin, 2) + 1, _format)
        else:
            binstr = format(self.value, _format)
        return binstr

    def decode(self, binary_str: str) -> int:
        """Populates the field value from binary and returns the next offset.
        
        Args:
            binary_str (str): The binary string to decode
        
        Returns:
            The bit offset after parsing
        """
        value = int(binary_str[:self.bits], 2)
        if (value & (1 << (self.bits - 1))) != 0:   #:sign bit set e.g. 8bit: 128-255
            value = value - (1 << self.bits)        #:compute negative value
        self.value = value
        return self.bits

    def xml(self) -> ET.Element:
        xmlfield = self._base_xml()
        size = ET.SubElement(xmlfield, 'Size')
        size.text = str(self.size)
        if self.default:
            default = ET.SubElement(xmlfield, 'Default')
            default.text = str(self.default)
        return xmlfield


class StringField(FieldCodec):
    """A character string sent over-the-air."""
    def __init__(self,
                 name: str,
                 size: int,
                 description: str = None,
                 optional: bool = False,
                 fixed: bool = False,
                 default: str = None,
                 value: str = None) -> None:
        """Instantiates a StringField.
        
        Args:
            name: The field name must be unique within a Message.
            size: The maximum number of characters in the string.
            description: An optional description/purpose for the string.
            optional: Indicates if the string is optional in the Message.
            fixed: Indicates if the string is always fixed length `size`.
            default: A default value for the string.
            value: Optional value to set during initialization.

        """
        super().__init__(name=name,
                         data_type='string',
                         description=description,
                         optional=optional)
        self.size = size
        self.fixed = fixed
        self.default = default
        self.value = value if value is not None else default
    
    def _validate_string(self, s: str):
        if s is not None:
            if not isinstance(s, str):
                raise ValueError(f'Invalid string {s}')
            if len(s) > self.size:
                warn(f'Clipping string at max {self.size} characters')
                return s[:self.size]
        return s
                
    @property
    def size(self):
        return self._size
    
    @size.setter
    def size(self, value: int):
        if not isinstance(value, int) or value < 1:
            raise ValueError('Size must be integer greater than 0 characters')
        self._size = value
    
    @property
    def default(self) -> str:
        return self._default
    
    @default.setter
    def default(self, v: str):
        self._default = self._validate_string(v)

    @property
    def value(self) -> str:
        return self._value
    
    @value.setter
    def value(self, v: str):
        self._value = self._validate_string(v)

    @property
    def bits(self):
        if self.fixed or self.value is None:
            return self.size * 8
        return len(self.value) * 8
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, StringField):
            return NotImplemented
        return _attribute_equivalence(self, other)

    def encode(self) -> str:
        """Returns the binary string of the field value."""
        if self.value is None and not self.optional:
            raise ValueError(f'No value defined for StringField {self.name}')
        binstr = ''.join(format(ord(c), '08b') for c in self.value)
        if self.fixed:
            binstr += ''.join('0' for bit in range(len(binstr), self.bits))
        else:
            binstr = _encode_field_length(len(self.value)) + binstr
        return binstr

    def decode(self, binary_str: str) -> int:
        """Populates the field value from binary and returns the next offset.
        
        Args:
            binary_str (str): The binary string to decode
        
        Returns:
            The bit offset after parsing
        """
        if self.fixed:
            length = self.size
            bit_index = 0
        else:
            (length, bit_index) = _decode_field_length(binary_str)
        n = int(binary_str[bit_index:bit_index + length * 8], 2)
        char_bytes = n.to_bytes((n.bit_length() + 7) // 8, 'big')
        for i, byte in enumerate(char_bytes):
            if byte == 0:
                warn('Truncating after 0 byte in string')
                char_bytes = char_bytes[:i]
                break
        self.value = char_bytes.decode('utf-8', 'surrogatepass') or '\0'
        return bit_index + length * 8

    def xml(self) -> ET.Element:
        """Returns the message definition XML representation of the field."""
        xmlfield = self._base_xml()
        size = ET.SubElement(xmlfield, 'Size')
        size.text = str(self.size)
        if self.fixed:
            fixed = ET.SubElement(xmlfield, 'Fixed')
            fixed.text = 'true'
        if self.default:
            default = ET.SubElement(xmlfield, 'Default')
            default.text = str(self.default)
        return xmlfield


class DataField(FieldCodec):
    """A data field of raw bytes sent over-the-air.
    
    Can also be used to hold floating point, double-precision or large integers.

    """
    supported_data_types = ['data', 'float', 'double']
    def __init__(self,
                 name: str,
                 size: int,
                 data_type: str = 'data',
                 description: str = None,
                 optional: bool = False,
                 fixed: bool = False,
                 default: bytes = None,
                 value: bytes = None) -> None:
        """Instantiates a EnumField.
        
        Args:
            name: The field name must be unique within a Message.
            size: The maximum number of bytes to send over-the-air.
            data_type: The data type represented within the bytes.
            description: An optional description/purpose for the field.
            optional: Indicates if the field is optional in the Message.
            fixed: Indicates if the data bytes are a fixed `size`.
            default: A default value for the enum.
            value: Optional value to set during initialization.

        """
        if data_type is None or data_type not in self.supported_data_types:
            raise ValueError(f'Invalid data type {data_type}')
        super().__init__(name=name,
                         data_type=data_type,
                         description=description,
                         optional=optional)
        self.fixed = fixed
        self.size = size
        self.default = default
        self.value = value if value is not None else default
    
    @property
    def size(self):
        return self._size
    
    @size.setter
    def size(self, value: int):
        if not isinstance(value, int) or value < 1:
            raise ValueError('Size must be integer greater than 0 bytes')
        if self.data_type == 'float':
            if value != 4:
                warn('Adjusting float size to 4 bytes fixed')
            self._size = 4
            self.fixed = True
        elif self.data_type == 'double':
            if value != 8:
                warn('Adjusting float size to 8 bytes fixed')
            self._size = 4
            self.fixed = True
        else:
            self._size = value
    
    @property
    def value(self):
        if self.data_type == 'float':
            return unpack('!f', self._value)
        if self.data_type == 'double':
            return unpack('!d', self._value)
        return self._value

    @value.setter
    def value(self, v: 'bytes|float'):
        if v is not None:
            if self.data_type in ['float', 'double']:
                if not isinstance(v, float):
                    raise ValueError(f'Invalid {self.data_type} value {v}')
                _format = '!f' if self.data_type == 'float' else '!d'
                v = pack(_format, v)
            elif not isinstance(v, bytes):
                raise ValueError(f'Invalid bytes {v}')
        self._value = v

    @property
    def bits(self):
        if self.fixed:
            return self.size * 8
        elif self.value is None:
            return 0
        return len(self.value) * 8
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DataField):
            return NotImplemented
        return _attribute_equivalence(self, other)

    def encode(self) -> str:
        """Returns the binary string of the field value."""
        if self.value is None and not self.optional:
            raise ValueError(f'No value defined for DataField {self.name}')
        binstr = ''
        binstr = ''.join(format(b, '08b') for b in self._value)
        if self.fixed:   #:pad to fixed length
            binstr += ''.join('0' for bit in range(len(binstr), self.bits))
        else:
            binstr = _encode_field_length(len(self._value)) + binstr
        return binstr

    def decode(self, binary_str: str) -> int:
        """Populates the field value from binary and returns the next offset.
        
        Args:
            binary_str (str): The binary string to decode
        
        Returns:
            The bit offset after parsing
        """
        if self.fixed:
            binary = binary_str[:self.bits]
            bits = self.bits
        else:
            (length, bit_index) = _decode_field_length(binary_str)
            binary = binary_str[bit_index:length * 8 + bit_index]
            bits = len(binary)
        self._value = int(binary, 2).to_bytes(int(bits / 8), 'big')
        return self.bits

    def xml(self) -> ET.Element:
        xmlfield = self._base_xml()
        size = ET.SubElement(xmlfield, 'Size')
        size.text = str(self.size)
        if self.default:
            default = ET.SubElement(xmlfield, 'Default')
            default.text = str(self.default)
        return xmlfield


class ArrayField(FieldCodec):
    """An Array Field provides a list where each element is a set of Fields.
    
    Attributes:
        name (str): The name of the field instance.
        size (int): The maximum number of elements allowed.
        fields (Fields): A list of Field types comprising each ArrayElement
        description (str): An optional description of the array/use.
        optional (bool): Indicates if the array is optional in the Message
        fixed (bool): Indicates if the array is always the fixed `size`
        elements (list): The enumerated list of ArrayElements

    """
    def __init__(self,
                 name: str,
                 size: int,
                 fields: Fields,
                 description: str = None,
                 optional: bool = False,
                 fixed: bool = False,
                 elements: "list[Fields]" = None) -> None:
        """Initializes an ArrayField instance.
        
        Args:
            name: The unique field name within the Message.
            size: The maximum number of elements allowed.
            fields: The list of Field types comprising each element.
            description: An optional description/purpose of the array.
            optional: Indicates if the array is optional in the Message.
            fixed: Indicates if the array is always the fixed `size`.
            elements: Option to populate elements of Fields during instantiation.

        """
        super().__init__(name=name,
                         data_type='array',
                         description=description,
                         optional=optional)
        self.size = size
        self.fixed = fixed
        self.fields = fields
        self.elements = elements or []
    
    @property
    def size(self):
        return self._size
    
    @size.setter
    def size(self, value: int):
        if not isinstance(value, int) or value < 1:
            raise ValueError('Size must be integer greater than 0 fields')
        self._size = value
    
    @property
    def fields(self):
        return self._fields

    @fields.setter
    def fields(self, fields: Fields):
        if not isinstance(fields, Fields):
            raise ValueError('Invalid Fields definition for ArrayField')
        self._fields = fields

    @property
    def elements(self):
        return self._elements
    
    @elements.setter
    def elements(self, elements: list):
        if elements is None or not hasattr(self, '_elements'):
            self._elements = []
        if not isinstance(elements, list):
            raise ValueError('Elements must be a list of grouped Fields')
        for fields in elements:
            for index, field in enumerate(fields):
                if (field.name != self.fields[index].name):
                    raise ValueError(f'fields[{index}].name'
                                     f' expected {self.fields[index].name}'
                                     f' got {field.name}')
                if (field.data_type != self.fields[index].data_type):
                    raise ValueError(f'fields[{index}].data_type'
                                     f' expected {self.fields[index].data_type}'
                                     f' got {field.data_type}')
                #TODO: validate non-optional fields have value/elements
                if field.value is None and not field.optional:
                    raise ValueError(f'fields[{index}].value missing')
                try:
                    self._elements[index] = fields
                except IndexError:
                    self._elements.append(fields)

    @property
    def bits(self):
        bits = 0
        for field in self.fields:
            bits += field.bits
        return bits
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ArrayField):
            return NotImplemented
        return _attribute_equivalence(self, other)

    def _valid_element(self, element: Fields) -> bool:
        for i, field in enumerate(self.fields):
            e = element[i]
            if e.name != field.name:
                raise ValueError(f'element field name {e.name}'
                                 f' does not match {field.name}')
            if e.data_type != field.data_type:
                raise ValueError(f'element field data_type {e.data_type}'
                                 f' does not match {field.data_type}')
            if e.optional != field.optional:
                raise ValueError(f'element optional {e.optional}'
                                 f' does not match {field.optional}')
            if hasattr(field, 'fixed') and e.fixed != field.fixed:
                raise ValueError(f'element fixed {e.fixed}'
                                 f' does not match {field.fixed}')
            if hasattr(field, 'size') and e.size != field.size:
                raise ValueError(f'element size {e.size}'
                                 f' does not match {field.size}')
        return True

    def append(self, element: Fields):
        """Adds the array element to the list of elements."""
        if not isinstance(element, Fields):
            raise ValueError('Invalid element definition must be Fields')
        if self._valid_element(element):
            for i, field in enumerate(element):
                if (hasattr(field, 'description') and
                    field.description != self.fields[i].description):
                    field.description = self.fields[i].description
                if hasattr(field, 'value') and field.value is None:
                    field.value = self.fields[i].default
            self._elements.append(element)

    def new_element(self) -> Fields:
        """Returns an empty element at the end of the elements list."""
        new_index = len(self._elements)
        self._elements.append(Fields(self.fields))
        return self.elements[new_index]

    def encode(self) -> str:
        """Returns the binary string of the field value."""
        if len(self.elements) == 0:
            raise ValueError('No elements to encode')
        binstr = ''
        for element in self.elements:
            for field in element:
                binstr += field.encode()
        if not self.fixed:
            binstr = _encode_field_length(len(self.elements)) + binstr
        return binstr

    def decode(self, binary_str: str) -> int:
        """Populates the field value from binary and returns the next offset.
        
        Args:
            binary_str (str): The binary string to decode
        
        Returns:
            The bit offset after parsing
        """
        if self.fixed:
            length = self.size
            bit_index = 0
        else:
            (length, bit_index) = _decode_field_length(binary_str)
        for index in range(0, length):
            fields = Fields(self.fields)
            for field in fields:
                if field.optional:
                    if binary_str[bit_index] == '0':
                        bit_index += 1
                        continue
                    bit_index += 1
                bit_index += field.decode(binary_str[bit_index:])
            try:
                self._elements[index] = fields
            except IndexError:
                self._elements.append(fields)
        return bit_index

    def xml(self) -> ET.Element:
        xmlfield = self._base_xml()
        size = ET.SubElement(xmlfield, 'Size')
        size.text = str(self.size)
        if self.fixed:
            default = ET.SubElement(xmlfield, 'Fixed')
            default.text = 'true'
        fields = ET.SubElement(xmlfield, 'Fields')
        for field in self.fields:
            fields.append(field.xml())
        return xmlfield


class MessageDefinitions:
    """A set of Message Definitions grouped into Services.

    Attributes:
        services: The list of Services with Messages defined.
    
    """
    def __init__(self, services: Services = None):
        if services is not None:
            if not isinstance(services, Services):
                raise ValueError('Invalid Services')
        self.services = services or Services()
    
    def xml(self, indent: bool = False) -> ET.Element:
        xmsgdef = ET.Element('MessageDefinition',
                             attrib={'xmlns:xsd': XML_NAMESPACE['xsd']})
        services = ET.SubElement(xmsgdef, 'Services')
        for service in self.services:
            services.append(service.xml())
        return xmsgdef if not indent else _indent_xml(xmsgdef)
    
    def mdf_export(self, filename: str, pretty: bool = False):
        tree = ET.ElementTree(self.xml())
        root = tree.getroot()
        if pretty:
            from xml.dom.minidom import parseString
            xmlstr = parseString(ET.tostring(root)).toprettyxml(indent="  ")
            with open(filename, 'w') as f:
                f.write(xmlstr)
        else:
            with open(filename, 'wb') as f:
                tree.write(f, encoding='utf-8', xml_declaration=True)



# -*- coding: utf-8 -*-
"""AT command protocol (threaded) for Inmarsat IDP satellite messaging modems.

This module provides a threaded serial interface that sends and receives 
AT commands, decoding/abstracting typically used operations.
Based on the PySerial threaded protocol factory, using a byte reader.

"""

import logging
import queue
import threading
from time import sleep, time

from serial import Serial, SerialException
from serial.threaded import LineReader, ReaderThread, Protocol

from idpmodem.aterror import AtCrcError, AtTimeout
from idpmodem.crcxmodem import get_crc, validate_crc


_log = logging.getLogger(__name__)


class AtProtocol(LineReader):
    """Threaded protocol factory for the IDP Modem.
    
    Maintains queues for requests, responses and unsolicited reports using
    task objects with callbacks.

    Accepts only one AT command at a time.  Handles command echo 
    terminated with <cr>, verbose response framed with <cr><lf>, 
    unsolicited data terminated with <lf>, and CRC error checking.

    Attributes:
        alive (bool): True while the factory is running.
        crc (bool): Indicates if CRC error checking is enabled.
        pending_command (str): The AT command being processed.
        responses (Queue): Queued responses to be processed as a line.
        unsolicited (Queue): Unexpected data received if no pending command.
        event_callback (Callable): optional callback function for 
        unexpected data
    """

    TERMINATOR = '\r'
    ENCODING = 'utf-8'
    UNICODE_HANDLING = 'replace'
    DEFAULT_TIMEOUT = 5

    def __init__(self,
                 event_callback: callable = None,
                 at_timeout: int = 5):
        """Initialize with CRC and optional callback

        Args:
            crc: Use CRC error checking (for long serial line).
            event_callback: Handler for non-command data.

        """
        super().__init__()
        # self.buffer = bytearray()   #: inherited from LineReader/Packetizer
        # self.transport = None   #: inherited from LineReader/Packetizer
        self.crc = False   #: This will be inferred from communications
        self.alive = True
        self.at_timeout = at_timeout
        self.pending_command = None
        self.command_time = None
        self.response_time = None
        self.responses = queue.Queue()
        self.events = queue.Queue()
        self._event_thread = threading.Thread(target=self._run_event,
                                              name='at_unsolicited',
                                              daemon=True)
        self._event_thread.start()
        self._lock = threading.Lock()
        self.event_callback = event_callback

    # def connection_made(self, transport):
    #     """Store transport - inherited from LineReader/Packetizer."""
    #     self.transport = transport

    # def connection_lost(self, exc):
    #     """Forget transport - inherited from LineReader/Packetizer."""
    #     self.transport = None
    #     if isinstance(exc, Exception):
    #         raise exc

    def data_received(self, data: bytearray):
        """Buffer received data and create packets for handlers.

        handle_packet() is inherited from LineReader.

        Args:
            data: a data byte received from the serial device

        """
        self.buffer.extend(data)
        if self.pending_command is not None:
            if data == b'\r':
                #: Echo case
                if self.buffer == self.pending_command.encode() + b'\r':
                    echo = self.buffer
                    #: reset buffer for response
                    self.buffer = bytearray(b'')
                    self.handle_packet(echo)
            elif data == b'\n':
                if (self.buffer != bytearray(b'\r\n')
                    and self.buffer != bytearray(b'\n')):
                    #: Framed/multiline response, error code or CRC
                    packet = self.buffer
                    self.buffer = bytearray(b'')
                    self.handle_packet(packet)
                elif self.buffer == bytearray(b'\n'):
                    #: (Unexpected) drop any empty lines
                    self.buffer = bytearray(b'')
        else:   # no pending command
            if data == b'\n':
                unsolicited = self.buffer
                self.buffer = bytearray(b'')
                self.handle_packet(unsolicited)

    # def handle_packet(self, packet: bytearray):
    #     """Decodes packet(s) - inherited from LineReader."""
    #     self.handle_line(packet.decode(self.ENCODING, self.UNICODE_HANDLING))

    def handle_line(self, line: str):
        """Enqueues lines for parsing by command handler.

        Args:
            line: The unicode string received from the serial port.
        """
        if self.pending_command is not None:
            self.responses.put(line)
        else:
            if line != '\n':
                self.events.put(line)

    def write_line(self, text):
        """Appends a terminator, encodes and writes text to the transport."""
        send = text + self.TERMINATOR
        self.transport.write(send.encode(self.ENCODING, self.UNICODE_HANDLING))
    
    def stop(self):
        """Stop the event processing thread and abort pending commands."""
        self.alive = False
        self.events.put(None)
        self.responses.put('<exit>')

    def _run_event(self):
        """Process events (unsolicited messages) in a separate thread.
        
        Ensures the command response thread is not blocked.

        Raises:
            AtUnsolicited: If an unexpected handling error occurs.
        
        """
        while self.alive:
            try:
                self.handle_event(self.events.get())
            except:
                _log.exception(f'Unexpected _run_event error')

    def handle_event(self, unsolicited: str):
        """Calls a user-defined function with the unicode string.

        Args:
            unsolicited: A unicode string terminated by <lf>.

        """
        if unsolicited is not None:
            if self.event_callback is not None:
                self.event_callback(unsolicited)
            else: 
                unsolicited = unsolicited.replace('\r', '<cr>')
                unsolicited = unsolicited.replace('\n', '<lf>')
                _log.warning(f'Unhandled event: {unsolicited}')

    def _clean_response(self,
                        lines: 'list[str]',
                        filter: 'list[str]' = [],
                        debug: bool = False,
                        ) -> 'list[str]':
        """Removes empty lines from response and returns a list.
        
        Args:
            lines: A list of reponse lines.
            filter: Optional list of strings/substrings to filter from response.
            debug: If True, logs the command latency
        
        Returns:
            List with filtered and stripped lines

        Raises:
            ValueError if filter is invalid.

        """
        if filter and not isinstance(filter, list):
            raise ValueError('filter must be a list of strings')
        if debug:
            latency = round(self.response_time - self.command_time, 3)
            _log.debug(f'Command {self.pending_command}'
                       f' latency: {latency} seconds')
        for l in range(len(lines)):
            lines[l] = lines[l].strip()
            if isinstance(filter, list) and len(filter) > 0:
                for s in filter:
                    if lines[l].startswith(s):
                        s += ':' if lines[l].startswith(f'{s}:') else ''
                        lines[l] = lines[l].replace(s, '').strip()
        return [x for x in lines if x != '']

    def command(self,
                command: str,
                filter: 'list[str]' = [],
                timeout: int = 5,
                debug: bool = False,
                ) -> 'list[str]':
        """Send an AT command and wait for the response.

        Returns the response as a list.  If an error response code was
        received then 'ERROR' will be the only string in the list.

        .. todo: generalize for OK-only response, and provide error detail

        Args:
            command: The AT command
            timeout: Time to wait for response in seconds (default 5)
        
        Returns:
            A list of strings. The list will be ['ERROR'] in case of a problem.

        Raises:
            AtCrcError if CRC does not match.
            AtTimeout if the request timed out.

        """
        with self._lock:  # ensure that just one thread is sending commands at once
            timeout = 1 if timeout < 1 else timeout
            command = get_crc(command) if self.crc else command
            self.pending_command = command
            self.command_time = time()
            self.response_time = None
            self.write_line(command)
            lines = []
            while self.pending_command is not None:
                try:
                    line: str = self.responses.get(timeout=timeout)
                    content = line.strip()
                    if self.response_time is None:
                        self.response_time = time()
                    if content == command:
                        pass   # ignore echo
                    elif content == 'OK':
                        lines.append(line)
                        if self.crc and '%CRC=0' in self.pending_command:
                            _log.debug('CRC disabled by command')
                            self.crc = False
                        if (not self.crc and
                            '%CRC=1' not in self.pending_command):
                            # No need to wait for CRC
                            return self._clean_response(lines, filter, debug)
                    elif content.startswith('*'):
                        if not self.crc:
                            if not '%CRC=1' in self.pending_command:
                                _log.debug('Found CRC enabled')
                            self.crc = True
                        crc = content.replace('*', '')
                        if not validate_crc(''.join(lines), crc):
                            raise AtCrcError(f'INVALID_CRC_RESPONSE')
                        return self._clean_response(lines, filter, debug)
                    else:   #: including 'ERROR'
                        lines.append(line)
                        # keep parsing in case CRC follows
                except queue.Empty:
                    if not self.response_time:
                        raise AtTimeout(f'TIMEOUT ({command})')
                    return self._clean_response(lines, filter, debug)


class ByteReaderThread(ReaderThread):
    """Modifies the ReaderThread class to process bytes individually.
    
    This is required due to complexities of optional checksum use 
    for long serial lines.

    """
    def __init__(self,
                 serial_instance: Serial,
                 protocol_factory: Protocol,
                 **kwargs):
        super().__init__(serial_instance, protocol_factory)
        self.name = None
        self.kwargs = kwargs

    def run(self):
        """Reader loop"""
        self.name = f'bytereader@{self.serial.name}'
        if not hasattr(self.serial, 'cancel_read'):
            self.serial.timeout = 1
        self.protocol = self.protocol_factory(**self.kwargs)
        try:
            self.protocol.connection_made(self)
        except Exception as err:
            self.alive = False
            self.protocol.connection_lost(err)
            self._connection_made.set()
            return
        error = None
        self._connection_made.set()
        data = bytearray()
        while self.alive and self.serial.is_open:
            try:
                # read all that is there or wait for one byte (blocking)
                if self.serial.in_waiting > 0:
                    data = self.serial.read()
                    self.protocol.data_received(data)
                sleep(0.001)
            except SerialException as err:
                # probably some I/O problem such as disconnected USB serial
                # adapters -> exit
                error = err
                break
            except Exception as err:
                error = err
                break
        self.alive = False
        self.protocol.connection_lost(error)
        self.protocol = None
